import os.path

import numpy as np
from numpy.testing import assert_almost_equal, assert_array_almost_equal, assert_array_equal
import pytest
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import train_test_split

from survivalist.datasets import load_gbsg2
from survivalist.exceptions import NoComparablePairException
from survivalist.functions import StepFunction
from survivalist.linear_model import CoxPHSurvivalAnalysis
from survivalist.metrics import (
    as_concordance_index_ipcw_scorer,
    as_cumulative_dynamic_auc_scorer,
    as_integrated_brier_score_scorer,
    brier_score,
    concordance_index_censored,
    concordance_index_ipcw,
    cumulative_dynamic_auc,
    integrated_brier_score,
)
from survivalist.nonparametric import kaplan_meier_estimator
from survivalist.preprocessing import OneHotEncoder
from survivalist.svm import FastSurvivalSVM
from survivalist.testing import FixtureParameterFactory, assert_cindex_almost_equal
from survivalist.util import Surv


def whas500_pred():
    WHAS500_DATA_FILE = os.path.join(os.path.dirname(__file__), "data", "whas500_predictions.csv")

    dat = np.loadtxt(WHAS500_DATA_FILE, delimiter=",")
    event = dat[:, 0] == 1
    time = dat[:, 1]
    risk = dat[:, 2]
    return event, time, risk


@pytest.fixture()
def no_comparable_pairs():
    y = np.array(
        [
            (False, 849.0),
            (False, 28.0),
            (False, 55.0),
            (False, 727.0),
            (False, 505.0),
            (False, 1558.0),
            (False, 1292.0),
            (False, 1737.0),
            (False, 944.0),
            (False, 750.0),
            (False, 2513.0),
            (False, 472.0),
            (False, 2417.0),
            (False, 538.0),
            (False, 49.0),
            (False, 723.0),
            (True, 3563.0),
            (False, 1090.0),
            (False, 1167.0),
            (False, 587.0),
            (False, 1354.0),
            (False, 910.0),
            (False, 398.0),
            (False, 854.0),
            (False, 3534.0),
            (False, 280.0),
            (False, 183.0),
            (False, 883.0),
            (False, 32.0),
            (False, 144.0),
        ],
        dtype=[("event", bool), ("time", float)],
    )
    scores = np.random.randn(y.shape[0])
    return y, scores


class ConcordanceIndexCases(FixtureParameterFactory):
    @property
    def time(self):
        return [1, 5, 6, 11, 34, 45, 46, 50]

    @property
    def estimate(self):
        return [5, 8, 11, 34, 12, 3, 9, 12]

    @property
    def time_tied(self):
        return [1, 5, 6, 11, 11, 34, 45, 45, 50]

    @property
    def estimate_tied(self):
        return [5, 8, 11, 19, 34, 12, 3, 9, 12]

    def data_no_censoring_all_correct(self):
        time = self.time
        event = np.ones(len(time), dtype=bool)
        estimate = np.arange(len(time))[::-1]
        expected = [1.0, 28, 0, 0, 0]
        return event, time, estimate, expected

    def data_no_censoring_all_wrong(self):
        time = self.time
        event = np.ones(len(time), dtype=bool)
        # order is exactly reversed
        estimate = np.arange(len(time))
        expected = [0.0, 0, 28, 0, 0]
        return event, time, estimate, expected

    def data_no_ties(self):
        event = [False, True, True, False, False, True, False, False]
        expected = [0.2307692, 3, 10, 0, 0]
        return event, self.time, self.estimate, expected

    def data_with_tied_time(self):
        event = [False, True, True, False, True, False, True, False, False]
        expected = [0.4, 8, 12, 0, 2]
        return event, self.time_tied, self.estimate_tied, expected

    def data_with_tied_time2(self):
        event = [False, True, True, False, False, False, True, False, False]
        expected = [0.2, 3, 12, 0, 1]
        return event, self.time_tied, self.estimate_tied, expected

    def data_with_tied_event(self):
        event = [False, True, False, True, True, False, True, False, False]
        expected = [0.5294118, 9, 8, 0, 1]
        return event, self.time_tied, self.estimate_tied, expected

    def data_with_tied_risk(self):
        event = [False, True, True, False, True, True, False, False]
        estimate = [5, 15, 11, 34, 12, 3, 9, 12]
        expected = [0.59375, 9, 6, 1, 0]
        return event, self.time, estimate, expected

    def data_with_almost_tied_risk(self):
        event, time, estimate, expected = self.data_with_tied_risk()
        estimate[4] += 4.5e-9
        estimate[-1] -= 4.5e-9
        return event, time, estimate, expected

    def data_with_tied_event_and_time(self):
        event = [True, False, False, False, True, False, True, True, False, False, False, True]
        time = [34, 11, 11, 5, 1, 89, 13, 45, 7, 13, 9, 13]
        estimate = [1, 19, 13, 13, 15, 14, 19, 23, 11, 10, 11, 1]

        expected = [0.5681818, 12, 9, 1, 2]
        return event, time, estimate, expected

    def data_whas500(self):
        event, time, estimate = whas500_pred()
        expected = [0.7697907, 57849, 17300, 0, 14]
        return event, time, estimate, expected


@pytest.mark.parametrize("event,time,estimate,expected", ConcordanceIndexCases().get_cases())
def test_concordance_index(event, time, estimate, expected):
    assert_cindex_almost_equal(event, time, estimate, expected)


class ConcordanceIndexFailureCases(FixtureParameterFactory):
    @property
    def event(self):
        return np.array([True, False, False, True, True, False])

    @property
    def time(self):
        return np.array([1, 5, 10, 12, 7, 65])

    @property
    def estimate(self):
        return np.array([12, 8, 1, 89, 56, 13])

    @property
    def error_different_length(self):
        return r"Found input variables with inconsistent numbers of samples: .+"

    def data_different_length_0(self):
        return (self.event, self.time[:3], self.estimate), self.error_different_length

    def data_different_length_1(self):
        return (self.event, self.time, self.estimate[:3]), self.error_different_length

    def data_different_length_2(self):
        return (self.event[:3], self.time, self.estimate), self.error_different_length

    def data_different_length_3(self):
        return (self.event, self.time[:3], self.estimate[:3]), self.error_different_length

    def data_different_length_4(self):
        return (self.event[:3], self.time, self.estimate[:3]), self.error_different_length

    def data_different_length_5(self):
        return (self.event[:3], self.time[:3], self.estimate), self.error_different_length

    def data_boolean_event(self):
        event = np.array([1, 0, 0, 1, 1, 0])

        match = "only boolean arrays are supported as class labels for survival analysis.+"
        return (event, self.time, self.estimate), match

    def data_min_samples(self):
        event = np.array([False])
        time = np.array([10])
        estimate = np.array([12])

        match = "Need a minimum of two samples"
        return (event, time, estimate), match

    def data_all_censored(self):
        event = np.array([False, False])
        time = np.array([10, 12])
        estimate = np.array([12, 13])

        match = "All samples are censored"
        return (event, time, estimate), match

    def data_all_finite_event_indicator(self):
        event = self.event.tolist()
        event[2] = None
        event = np.array(event)

        match = "Input event_indicator contains NaN"
        return (event, self.time, self.estimate), match

    def data_all_finite_event_time(self):
        time = self.time.tolist()
        time[3] = np.nan
        time = np.array(time)

        match = "Input event_time contains NaN"
        return (self.event, time, self.estimate), match

    def data_all_finite_estimate(self):
        estimate = self.estimate.tolist()
        estimate[5] = np.inf
        estimate = np.array(estimate)

        match = "Input estimate contains infinity or a value too large"
        return (self.event, self.time, estimate), match

    def _get_not_1d(self, dim):
        event, time, risk = whas500_pred()

        risk = np.tile(risk[:, np.newaxis], (1, dim))

        match = "Expected 1D array, got 2D array instead:"
        return (event, time, risk), match

    def data_not_1d_dim2(self):
        return self._get_not_1d(2)

    def data_not_1d_dim3(self):
        return self._get_not_1d(3)

    def data_not_1d_dim10(self):
        return self._get_not_1d(10)


@pytest.mark.parametrize("inputs,match", ConcordanceIndexFailureCases().get_cases())
def test_concordance_index_failure(inputs, match):
    with pytest.raises(ValueError, match=match):
        concordance_index_censored(*inputs)


def test_concordance_index_no_comparable(no_comparable_pairs):
    y, scores = no_comparable_pairs

    with pytest.raises(NoComparablePairException):
        concordance_index_censored(y["event"], y["time"], scores)


class UnoCCases(FixtureParameterFactory):
    @property
    def y_train(self):
        return Surv.from_arrays(
            time=(2, 4, 6, 8, 10, 11, 15, 19),
            event=(False, True, False, True, False, False, False, False),
        )

    @property
    def y_ties(self):
        return Surv.from_arrays(
            event=np.array((0, 1, 1, 0, 1, 0, 1, 0, 0, 1), dtype=bool),
            time=(1, 5, 6, 10, 11, 34, 45, 46, 50, 56),
        )

    def data_no_ties(self):
        estimate = (5, 8, 11, 19, 34, 12, 3, 9, 12, 20)
        expected = (0.347890360332615, 8, 15, 0, 0)
        return self.y_ties, self.y_ties, estimate, expected, None

    def data_tied_risk_1(self):
        estimate = (5, 8, 11, 11, 34, 12, 3, 9, 12, 20)
        expected = (0.365629810028969, 8, 14, 1, 0)
        return self.y_ties, self.y_ties, estimate, expected, None

    def data_tied_risk_2(self):
        estimate = (5, 8, 11, 11, 34, 12, 11, 9, 12, 20)
        expected = (0.387865723332956, 7, 14, 2, 0)
        return self.y_ties, self.y_ties, estimate, expected, None

    def data_truncated_1(self):
        y_test = Surv.from_arrays(
            time=(1, 3, 5, 8, 12, 13),
            event=(True, False, False, True, True, True),
        )
        estimate = (5, 8, 13, 11, 9, 4)
        expected = (0.7543736528146774, 4, 4, 0, 0)
        tau = 19
        return self.y_train, y_test, estimate, expected, tau

    def data_truncated_2(self):
        y = Surv.from_arrays(
            time=(1, 5, 6, 10, 11, 34, 45, 46, 50, 56),
            event=np.array((0, 1, 1, 0, 1, 0, 1, 1, 1, 1), dtype=bool),
        )
        estimate = (5, 8, 11, 19, 34, 12, 3, 9, 12, 18)
        expected = (0.347890361949191, 8, 18, 0, 0)
        tau = 45.25
        return y, y, estimate, expected, tau

    def data_last_time_censored(self):
        y_test = Surv.from_arrays(
            time=(1, 3, 5, 7, 12, 13, 20),
            event=(True, False, False, True, True, False, False),
        )
        estimate = (5, 8, 13, 11, 9, 7, 4)
        expected = (0.8126567565914234, 6, 5, 0, 0)
        return self.y_train, y_test, estimate, expected, None

    def data_tied_event(self):
        y = Surv.from_arrays(
            event=[False, True, False, True, True, False, True, False, False, True],
            time=[1, 5, 6, 11, 11, 34, 45, 45, 50, 55],
        )
        estimate = (5, 8, 11, 19, 34, 12, 3, 9, 12, 18)
        expected = (0.4036321031048623, 11, 10, 0, 1)
        return y, y, estimate, expected, None

    def data_tied_event_and_time(self):
        y = Surv.from_arrays(
            event=[True, False, False, False, True, False, True, True, False, False, False, True, True],
            time=[34, 11, 11, 5, 1, 89, 13, 45, 7, 13, 9, 13, 90],
        )
        estimate = (1, 19, 13, 13, 15, 14, 19, 23, 11, 10, 11, 1, 18)
        expected = (0.46795357052737824, 14, 12, 1, 2)
        return y, y, estimate, expected, None

    def data_whas500(self):
        event, time, estimate = whas500_pred()
        y = Surv.from_arrays(event, time)
        expected = (0.7929275009049014, 57849, 17300, 0, 14)
        return y, y, estimate, expected, None


@pytest.mark.parametrize("y_train,y_test,estimate,expected,tau", UnoCCases().get_cases())
def test_uno_c(y_train, y_test, estimate, expected, tau):
    result = concordance_index_ipcw(y_train, y_test, estimate, tau=tau)
    assert_array_equal(result[1:], expected[1:])
    assert_almost_equal(result[0], expected[0])


class UnoCFailureCases(FixtureParameterFactory):
    @property
    def y_train(self):
        return Surv.from_arrays(
            time=(2, 4, 6, 8, 10, 11, 15, 19), event=(False, True, False, True, False, False, False, True)
        )

    @property
    def y_test(self):
        return Surv.from_arrays(time=(1, 3, 5, 7, 12, 13, 19), event=(True, False, False, True, True, False, True))

    @property
    def estimate(self):
        return (5, 8, 13, 11, 9, 7, 4)

    def data_last_time_uncensored_1(self):
        y_train = self.y_train
        y_test = self.y_test
        y_test["time"][-1] = 20
        match = "time must be smaller than largest observed time point:"

        inputs = (y_train, y_test, self.estimate)
        return inputs, match

    def data_last_time_uncensored_2(self):
        y_train = self.y_train
        y_test = Surv.from_arrays(time=(1, 23, 5, 27, 12), event=(True, False, True, True, False))
        estimate = (5, 13, 11, 9, 4)
        match = "time must be smaller than largest observed time point:"

        inputs = (y_train, y_test, estimate)
        return inputs, match

    def data_zero_prob_1(self):
        y_train = self.y_train
        y_train["event"][-1] = False
        match = "censoring survival function is zero at one or more time points"

        inputs = (y_train, self.y_test, self.estimate)
        return inputs, match

    def data_zero_prob_2(self):
        y_train = self.y_train
        y_train["time"][-1] = 18
        y_train["event"][-1] = False

        match = "censoring survival function is zero at one or more time points"

        inputs = (y_train, self.y_test, self.estimate)
        return inputs, match

    def data_zero_prob_3(self):
        y_train = self.y_train
        y_train["time"][-1] = 18
        y_train["event"][-1] = False

        y_test = self.y_test
        y_test["time"] = (1, 3, 5, 19, 12, 13, 7)
        match = "censoring survival function is zero at one or more time points"

        inputs = (y_train, y_test, self.estimate)
        return inputs, match

    def _get_not_1d(self, dim):
        event, time, risk = whas500_pred()
        y = Surv.from_arrays(event, time)

        risk = np.tile(risk[:, np.newaxis], (1, dim))

        match = "Expected 1D array, got 2D array instead:"
        inputs = (y, y, risk)
        return inputs, match

    def data_not_1d_dim2(self):
        return self._get_not_1d(2)

    def data_not_1d_dim3(self):
        return self._get_not_1d(3)

    def data_not_1d_dim10(self):
        return self._get_not_1d(10)


@pytest.mark.parametrize("inputs,expected_msg", UnoCFailureCases().get_cases())
def test_uno_c_failure(inputs, expected_msg):
    with pytest.raises(ValueError, match=expected_msg):
        concordance_index_ipcw(*inputs)


def test_uno_c_all_censored():
    y_train = Surv.from_arrays(
        time=(2, 4, 6, 8, 10, 11, 15, 19),
        event=(True, True, True, True, True, True, True, True),
    )
    y_test = Surv.from_arrays(
        time=(1, 3, 5, 7, 12, 13, 20),
        event=(True, False, False, True, True, False, False),
    )
    estimate = (5, 8, 13, 11, 9, 7, 4)

    ret_uno = concordance_index_ipcw(y_train, y_test, estimate)
    ret_harrell = concordance_index_censored(y_test["event"], y_test["time"], estimate)
    assert ret_uno == ret_harrell


def test_uno_c_no_comparable(no_comparable_pairs):
    y, scores = no_comparable_pairs

    with pytest.raises(NoComparablePairException):
        concordance_index_ipcw(y, y, scores)


class BaseUnoCAucCases(FixtureParameterFactory):
    @property
    def uno_auc_y(self):
        return Surv.from_arrays(
            time=[10.88, 19.78, 40.92, 98.7, 70.19, 10.15, 28.95, 29.57, 17.9, 63.78, 36.22, 83.14, 13.69, 99.51, 3.19],
            event=[1, 0, 1, 0, 1, 1, 1, 1, 1, 0, 0, 1, 0, 0, 1],
        )

    @property
    def uno_auc_data_15(self):
        estimate = [
            -1.019,
            -0.016,
            0.132,
            0.269,
            -0.777,
            -1.077,
            0.894,
            -1.227,
            -0.417,
            0.072,
            -1.275,
            -0.91,
            -0.825,
            -0.292,
            -0.045,
        ]
        return self.uno_auc_y, estimate

    @property
    def uno_auc_data_20(self):
        y_train = Surv.from_arrays(
            time=[
                77.6,
                57.6,
                66.6,
                67.0,
                31.5,
                5.5,
                67.4,
                43.7,
                31.7,
                71.9,
                81.1,
                56.2,
                88.1,
                2.9,
                62.0,
                17.2,
                88.0,
                26.4,
                93.5,
                79.9,
            ],
            event=[1, 1, 0, 1, 0, 1, 1, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1],
        )
        estimate = [
            -1.019,
            -0.016,
            0.132,
            0.269,
            -0.777,
            -1.077,
            0.894,
            -1.227,
            -0.417,
            0.072,
            -1.275,
            -0.91,
            -0.825,
            -0.292,
            -0.045,
        ]
        return y_train, self.uno_auc_y, estimate

    @property
    def uno_auc_time_dependent_20(self):
        y_train, y_test, _ = self.uno_auc_data_20

        estimate = np.array(
            [
                [
                    0.566,
                    0.576,
                    0.506,
                    1.871,
                    -0.04,
                    0.281,
                    0.335,
                    -1.181,
                    1.368,
                    -0.192,
                    0.955,
                    0.221,
                    0.057,
                    -0.996,
                    -1.27,
                ],
                [
                    -2.658,
                    -1.121,
                    1.903,
                    -0.7,
                    -1.013,
                    -0.472,
                    -0.668,
                    -0.537,
                    1.659,
                    0.657,
                    -1.317,
                    -1.103,
                    2.159,
                    0.625,
                    -0.067,
                ],
                [
                    -1.004,
                    -1.67,
                    0.775,
                    -2.512,
                    1.179,
                    0.073,
                    0.103,
                    -0.379,
                    -0.082,
                    -0.617,
                    1.287,
                    -0.449,
                    -0.477,
                    0.24,
                    0.838,
                ],
            ]
        ).T

        return y_train, y_test, estimate


class UnoCAucCases(BaseUnoCAucCases):
    def data_single_time(self):
        y_train, estimate = self.uno_auc_data_15
        times = 28
        iauc = 0.362963
        expected = np.array([iauc])

        return y_train, y_train, estimate, times, expected, iauc

    def data_two_times(self):
        y_train, estimate = self.uno_auc_data_15
        times = [15, 66]
        iauc = 0.3943949
        expected = np.array([0.3030303, 0.4500000])

        return y_train, y_train, estimate, times, expected, iauc

    def data_two_times_int(self):
        y_train, estimate = self.uno_auc_data_15
        y_train["time"] = y_train["time"] * 100.0
        y_train = y_train.astype([("event", bool), ("time", int)])
        estimate = (np.array(estimate) * 1000).astype(int)
        times = np.array([1500, 6600], dtype=int)
        iauc = 0.3943949
        expected = np.array([0.3030303, 0.4500000])

        return y_train, y_train, estimate, times, expected, iauc

    def data_min_to_max_times(self):
        y_train, estimate = self.uno_auc_data_15
        times = (y_train["time"].min(), 15, 66, y_train["time"].max() - 1e-6)
        iauc = 0.3999539
        expected = np.array([0.6428571, 0.3030303, 0.4500000, 0.3162996])

        return y_train, y_train, estimate, times, expected, iauc

    def data_train_test(self):
        y_train, y_test, estimate = self.uno_auc_data_20
        times = [15, 66]
        iauc = 0.385509
        expected = np.array([0.3030303, 0.4357061])

        return y_train, y_test, estimate, times, expected, iauc

    def data_tied_test_time(self):
        y_train, y_test, estimate = self.uno_auc_data_20
        y_test["time"][0] = y_test["time"][1]
        times = [15, 66]
        iauc = 0.4204885
        expected = np.array([0.3750000, 0.4357061])

        return y_train, y_test, estimate, times, expected, iauc

    def data_tied_test_score(self):
        y_train, y_test, estimate = self.uno_auc_data_20
        estimate[0] = estimate[-1]
        times = [15, 66]
        iauc = 0.495604291
        expected = np.array([0.4242424, 0.539036])

        return y_train, y_test, estimate, times, expected, iauc

    def data_tied_test_score_max(self):
        y_train, y_test, estimate = self.uno_auc_data_20
        o = np.argsort(estimate)
        estimate[o[0]] = estimate[o[1]]
        times = [15, 66]
        iauc = 0.385509
        expected = np.array([0.3030303, 0.4357061])

        return y_train, y_test, estimate, times, expected, iauc

    def data_tied_test_score_min(self):
        y_train, y_test, estimate = self.uno_auc_data_20
        o = np.argsort(estimate)
        estimate[o[-1]] = estimate[o[-2]]
        times = [15, 66]
        iauc = 0.374134
        expected = np.array([0.3030303, 0.4174082])

        return y_train, y_test, estimate, times, expected, iauc

    def data_time_dependent(self):
        y_train, y_test, estimate = self.uno_auc_time_dependent_20
        times = [15, 30, 72]
        iauc = 0.5522067
        expected = np.array([0.3636364, 0.5247813, 0.7603000])

        return y_train, y_test, estimate, times, expected, iauc

    def _compute_roc_auc(self, y, times, estimate):
        expected_auc = np.array([roc_auc_score(y["time"] > t, e) for t, e in zip(times, estimate)])
        km_delta = np.array([1 - 0.8, 0.8 - 0.5, 0.5 - 0.2])
        expected_iauc = np.sum(km_delta * expected_auc) / 0.8

        return expected_auc, expected_iauc

    @property
    def y_roc_auc(self):
        return Surv.from_arrays(
            time=[7, 9, 11, 12, 13, 15, 28, 39, 41, 76],
            event=[True, True, True, True, True, True, True, True, True, True],
        )

    @property
    def times_roc_auc(self):
        return [10, 14, 40]

    def data_time_dependent_without_censoring(self):
        estimate = np.array(
            [
                [1, 6, 18, 56, 32, 3, 99, 7, 67, 541],
                [6, 9, 11, 5, 3, 12, 56, 56.1, 81, 77],
                [13, 11, 12, 76, 55, 134, 70, 78, 75, 99],
            ]
        )
        y = self.y_roc_auc
        times = self.times_roc_auc
        expected_auc, expected_iauc = self._compute_roc_auc(y, times, estimate)

        return y, y, -estimate.T, times, expected_auc, expected_iauc

    def data_time_dependent_with_ties_without_censoring(self):
        estimate = np.array(
            [
                [1, 6, 7, 56, 32, 3, 99, 7, 79, 17],
                [3, 6, 11, 5, 17, 12, 17, 56.1, 81, 77],
                [13, 11, 12, 17, 17, 134, 70, 78, 13, 99],
            ]
        )
        y = self.y_roc_auc
        times = self.times_roc_auc
        expected_auc, expected_iauc = self._compute_roc_auc(y, times, estimate)

        return y, y, -estimate.T, times, expected_auc, expected_iauc

    def data_whas500(self):
        event, time, estimate = whas500_pred()
        y_train = Surv.from_arrays(event=event[:300], time=time[:300])
        y_test = Surv.from_arrays(event=event[300:], time=time[300:])
        estimate = estimate[300:]
        times = (200, 400, 600, 800, 1000, 1200, 1400)

        iauc = 0.8045058
        expected = np.array([0.7720669, 0.7765915, 0.7962623, 0.8759295, 0.8759295, 0.8759513, 0.9147647])
        return y_train, y_test, estimate, times, expected, iauc

    def data_whas500_unordered_time(self):
        y_train, y_test, estimate, _, expected, iauc = self.data_whas500()
        times = (1000, 600, 1400, 200, 400, 1200, 800, 1000, 200)
        return y_train, y_test, estimate, times, expected, iauc


@pytest.mark.parametrize("y_train,y_test,estimate,times,expect_auc,expect_iauc", UnoCAucCases().get_cases())
def test_uno_auc(y_train, y_test, estimate, times, expect_auc, expect_iauc):
    auc, iauc = cumulative_dynamic_auc(y_train, y_test, estimate, times)
    assert_array_almost_equal(auc, expect_auc)
    if isinstance(expect_iauc, np.ndarray):
        assert_almost_equal(iauc, expect_iauc)
    else:
        assert iauc == pytest.approx(expect_iauc)


class UnoCAucFailureCases(BaseUnoCAucCases):
    @property
    def times(self):
        return 33

    def data_time_too_big_1(self):
        y_train, y_test, _ = self.uno_auc_data_20
        y_test["time"][11] = 100
        match = "time must be smaller than largest observed time point:"

        return y_train, y_test, self.times, match

    def data__time_too_big_2(self):
        y_train, y_test, _ = self.uno_auc_data_20
        idx = np.argmax(y_test["time"])
        y_test["event"][idx] = True
        match = "time must be smaller than largest observed time point:"

        return y_train, y_test, self.times, match

    def data_ipcw_undefined_1(self):
        y_train, y_test, _ = self.uno_auc_data_20
        idx = np.argmax(y_train["time"])
        y_train["event"][idx] = 0
        y_test["time"][0] = y_train["time"][idx]
        match = "censoring survival function is zero at one or more time points"

        return y_train, y_test, self.times, match

    def data_ipcw_undefined_2(self):
        y_train, y_test, _ = self.uno_auc_data_20
        idx = np.argmax(y_train["time"])
        y_train["event"][idx] = 0
        y_test["time"][-1] = y_train["time"][idx] * 2
        y_test["event"][-1] = True
        match = "censoring survival function is zero at one or more time points"

        return y_train, y_test, self.times, match

    def data_all_censored_train(self):
        y_train, y_test, _ = self.uno_auc_data_20
        y_train["event"] = False
        match = "all samples are censored"

        return y_train, y_test, self.times, match

    def data_all_censored_test(self):
        y_train, y_test, _ = self.uno_auc_data_20
        y_test["event"] = False
        match = "all samples are censored"

        return y_train, y_test, self.times, match

    def data_times_nan(self):
        y_train, y_test, _ = self.uno_auc_data_20
        times = (0.2, np.nan)
        match = r"Input times contains NaN"

        return y_train, y_test, times, match

    def data_times_infinite_1(self):
        y_train, y_test, _ = self.uno_auc_data_20
        times = (0.2, np.inf)
        match = r"Input times contains infinity or a value too large for dtype\('float64'\)."

        return y_train, y_test, times, match

    def data_times_infinite_2(self):
        y_train, y_test, _ = self.uno_auc_data_20
        times = (0.2, -np.inf)
        match = r"Input times contains infinity or a value too large for dtype\('float64'\)."

        return y_train, y_test, times, match

    def data_times_too_big_1(self):
        y_train, y_test, _ = self.uno_auc_data_20
        idx = np.argmax(y_test["time"])
        y_test["event"][idx] = False
        t_max = y_test["time"][idx]
        times = (33, t_max / 2, t_max)
        match = r"all times must be within follow-up time of test data: \[3\.19; 99\.51\["

        return y_train, y_test, times, match

    def data_times_too_big_2(self):
        y_train, y_test, times, match = self.data_times_too_big_1()
        times = times[:-1] + (times[-1] + 0.1,)
        match = r"all times must be within follow-up time of test data: \[3\.19; 99\.51\["

        return y_train, y_test, times, match

    def data_times_too_big_3(self):
        y_train, y_test, _ = self.uno_auc_data_20
        max_train = np.max(y_train["time"])
        idx_test = y_test["time"] > max_train
        y_test["event"][idx_test] = False
        y_test["time"][idx_test] = max_train
        times = (33, max_train)
        match = r"all times must be within follow-up time of test data: \[3\.19; 93\.5\["

        return y_train, y_test, times, match

    def data_times_too_small_1(self):
        y_train, y_test, _ = self.uno_auc_data_20
        idx = np.argmin(y_test["time"])
        y_test["event"][idx] = True
        t_min = y_test["time"][idx]
        times = (t_min - 1e-6, 33)
        match = r"all times must be within follow-up time of test data: \[3\.19; 99\.51\["

        return y_train, y_test, times, match

    def data_times_too_small_2(self):
        y_train, y_test, _ = self.uno_auc_data_20
        idx = np.argmin(y_test["time"])
        y_test["event"][idx] = True
        t_min = y_test["time"][idx]
        times = (33, t_min - 0.1, t_min / 2)
        match = r"all times must be within follow-up time of test data: \[3\.19; 99\.51\["

        return y_train, y_test, times, match

    def data_times_empty(self):
        y_train, y_test, _ = self.uno_auc_data_20
        times = []
        match = r"Found array with 0 sample\(s\)"

        return y_train, y_test, times, match


@pytest.mark.parametrize("y_train,y_test,times,match", UnoCAucFailureCases().get_cases())
def test_uno_auc_failure(y_train, y_test, times, match):
    estimate = np.random.randn(y_test.shape[0])
    with pytest.raises(ValueError, match=match):
        cumulative_dynamic_auc(y_train, y_test, estimate, times)


class UnoAucShapeFailureCases(BaseUnoCAucCases):
    @property
    def times(self):
        return 33

    def data_estimate_2d_1col(self):
        y_train, y_test, estimate = self.uno_auc_data_20
        estimate = np.atleast_2d(estimate)
        match = r"Found input variables with inconsistent numbers of samples: \[15, 1\]"

        return y_train, y_test, self.times, estimate, match

    def data_estimate_2d_2cols(self):
        y_train, y_test, estimate = self.uno_auc_data_20
        times = [11, 33, 55]
        estimate = np.tile(estimate, (2, 1)).T
        match = "expected estimate with 3 columns, but got 2"

        return y_train, y_test, times, estimate, match

    def data_estimate_2d_4cols(self):
        y_train, y_test, estimate = self.uno_auc_data_20
        times = [11, 33, 55]
        estimate = np.tile(estimate, (4, 1)).T
        match = "expected estimate with 3 columns, but got 4"

        return y_train, y_test, times, estimate, match

    def data_estimate_3d(self):
        y_train, y_test, estimate = self.uno_auc_data_20
        estimate = np.atleast_3d(estimate)
        match = "Found array with dim 3. cumulative_dynamic_auc expected <= 2."

        return y_train, y_test, self.times, estimate, match


@pytest.mark.parametrize("y_train,y_test,times,estimate,match", UnoAucShapeFailureCases().get_cases())
def test_uno_auc_shape_failure(y_train, y_test, times, estimate, match):
    with pytest.raises(ValueError, match=match):
        cumulative_dynamic_auc(y_train, y_test, estimate, times)


@pytest.fixture()
def nottingham_prognostic_index():
    def _get_npi(times):
        X, y = load_gbsg2()

        grade = X.loc[:, "tgrade"].map({"I": 1, "II": 2, "III": 3}).astype(int)
        NPI = 0.2 * X.loc[:, "tsize"] / 10 + 1 + grade
        NPI[NPI < 3.4] = 1.0
        NPI[(NPI >= 3.4) & (NPI <= 5.4)] = 2.0
        NPI[NPI > 5.4] = 3.0

        preds = np.empty((X.shape[0], len(times)), dtype=float)
        for j, ts in enumerate(times):
            survs = {}
            for i in NPI.unique():
                idx = np.flatnonzero(NPI == i)
                yi = y[idx]
                t, s = kaplan_meier_estimator(yi["cens"], yi["time"])
                if t[-1] < ts and s[-1] == 0.0:
                    survs[i] = 0.0
                else:
                    fn = StepFunction(t, s)
                    survs[i] = fn(ts)

            preds[:, j] = NPI.map(survs).values

        return preds, y

    return _get_npi


@pytest.fixture(params=[365, 730, 1095, 1460, 1825])
def brier_npi_data(request, nottingham_prognostic_index):
    t = request.param

    pred, y = nottingham_prognostic_index([t])

    if t == 365:
        bs = 0.0762922458520448
    elif t == 730:
        bs = 0.182536421174199
    elif t == 1095:
        bs = 0.220017747254941
    elif t == 1460:
        bs = 0.234133800146671
    elif t == 1825:
        bs = 0.233822955042198
    else:
        raise AssertionError()

    return pred, y, t, bs


def test_brier_nottingham(brier_npi_data):
    pred, y, times, expected_score = brier_npi_data

    _, score = brier_score(y, y, pred.squeeze(), times=times)
    assert score[0] == pytest.approx(expected_score)


def test_brier_nottingham_many(nottingham_prognostic_index):
    times = [365, 730, 1095, 1460, 1825]
    pred, y = nottingham_prognostic_index(times)

    expected_score = np.array(
        [
            0.0762922458520448,
            0.182536421174199,
            0.220017747254941,
            0.234133800146671,
            0.233822955042198,
        ]
    )

    t1, score = brier_score(y, y, pred.squeeze(), times=times)
    assert_array_almost_equal(score, expected_score)

    t2, score = brier_score(y, y, pred.squeeze(), times=times[::-1])
    assert_array_almost_equal(score, expected_score)
    assert_array_equal(t1, t2)


def test_brier_times_too_large(nottingham_prognostic_index):
    pred, y = nottingham_prognostic_index([1825])

    with pytest.raises(ValueError, match="all times must be within follow-up time of test data:"):
        brier_score(y, y, pred, times=9999)


def test_brier_wrong_estimate_shape(nottingham_prognostic_index):
    pred, y = nottingham_prognostic_index([720, 1825])

    with pytest.raises(ValueError, match="expected estimate with 2 columns, but got 1"):
        brier_score(y, y, pred[:, :1], times=[720, 1825])

    with pytest.raises(ValueError, match="expected estimate with 3 columns, but got 2"):
        brier_score(y, y, pred, times=[720, 960, 1825])

    with pytest.raises(ValueError, match=r"Found input variables with inconsistent numbers of samples: \[686, 10\]"):
        brier_score(y, y, pred[:10], times=[720, 1825])


def test_brier_coxph():
    X, y = load_gbsg2()
    X["tgrade"] = X.loc[:, "tgrade"].map(len).astype(int)

    Xt = OneHotEncoder().fit_transform(X)

    survs = CoxPHSurvivalAnalysis(ties="efron").fit(Xt, y).predict_survival_function(Xt)

    preds = [fn(1825) for fn in survs]

    _, score = brier_score(y, y, preds, 1825)

    assert score[0] == pytest.approx(0.208817407492645, 1e-5)


def test_brier_score_int_dtype():
    times = np.arange(1, 31, dtype=int)
    rnd = np.random.RandomState(1)
    times = rnd.choice(times, 20)

    y_int = np.empty(20, dtype=[("event", bool), ("time", int)])
    y_int["event"] = np.ones(20, dtype=bool)
    y_int["event"][:10] = False
    y_int["time"] = times

    pred = rnd.randn(20, 10)
    tp = np.linspace(1.0, 2.0, 10)
    _, bs_int = brier_score(y_int, y_int, pred, times=tp)

    y_float = np.empty(20, dtype=[("event", bool), ("time", float)])
    y_float["event"][:] = y_int["event"]
    y_float["time"][:] = y_int["time"]
    _, bs_float = brier_score(y_float, y_float, pred, times=tp)

    assert_array_almost_equal(bs_float, bs_int)


def test_ibs_nottingham_1(nottingham_prognostic_index):
    times = np.linspace(365, 1825, 5)  # t=1..5 years
    preds, y = nottingham_prognostic_index(times)

    score = integrated_brier_score(y, y, preds, times=times)
    assert score == pytest.approx(0.197936392255733)

    score = integrated_brier_score(y, y, preds[:, :4], times=times[:4])
    assert score == pytest.approx(0.185922397142833)


def test_ibs_nottingham_2(nottingham_prognostic_index):
    times = np.arange(1095, 1826)  # t=3..5 years
    preds, y = nottingham_prognostic_index(times)

    score = integrated_brier_score(y, y, preds, times=times)

    assert score == pytest.approx(0.231553687189643)


def test_ibs_single_time_point(nottingham_prognostic_index):
    pred, y = nottingham_prognostic_index([1825])

    with pytest.raises(ValueError, match="At least two time points must be given"):
        integrated_brier_score(y, y, pred, times=1825)


@pytest.fixture(params=["cindex", "auc", "brier"])
def scorers_data(request, make_whas500):
    t = request.param

    whas500_data = make_whas500(to_numeric=True)
    data = train_test_split(whas500_data.x, whas500_data.y, random_state=0, stratify=whas500_data.y["fstat"])
    times = np.percentile(whas500_data.y["lenfol"], np.linspace(5, 81, 15))

    if t == "cindex":

        def func(*args, **kwargs):
            ret = concordance_index_ipcw(*args, **kwargs)
            return ret[0]

        wrapper_cls = as_concordance_index_ipcw_scorer
        args = {}
    elif t == "auc":

        def func(*args, **kwargs):
            ret = cumulative_dynamic_auc(*args, **kwargs)
            return ret[1]

        wrapper_cls = as_cumulative_dynamic_auc_scorer
        args = {"times": times}
    elif t == "brier":
        func = integrated_brier_score
        wrapper_cls = as_integrated_brier_score_scorer
        args = {"times": times}
    else:
        raise AssertionError()

    return func, wrapper_cls, args, data


def test_scorers(scorers_data):
    score_func, wrapper_cls, score_args, data = scorers_data
    X_train, X_test, y_train, y_test = data

    est_std = CoxPHSurvivalAnalysis().fit(X_train, y_train)
    if issubclass(wrapper_cls, as_integrated_brier_score_scorer):
        times = score_args["times"]
        pred = np.vstack([fn(times) for fn in est_std.predict_survival_function(X_test)])
        sign = -1
    else:
        pred = est_std.predict(X_test)
        sign = 1

    expected = sign * score_func(y_train, y_test, pred, **score_args)

    est_wrap = wrapper_cls(CoxPHSurvivalAnalysis(), **score_args)
    est_wrap.fit(X_train, y_train)
    actual = est_wrap.score(X_test, y_test)

    assert actual == pytest.approx(expected)

    assert_array_almost_equal(est_wrap.predict(X_test), est_std.predict(X_test))

    chf_expected = est_std.predict_cumulative_hazard_function(X_test)
    chf_actual = est_wrap.predict_cumulative_hazard_function(X_test)
    assert_array_almost_equal([v.x for v in chf_expected], [v.x for v in chf_actual])
    assert_array_almost_equal([v.y for v in chf_expected], [v.y for v in chf_actual])

    surv_expected = est_std.predict_survival_function(X_test)
    surv_actual = est_wrap.predict_survival_function(X_test)
    assert_array_almost_equal([v.x for v in surv_expected], [v.x for v in surv_actual])
    assert_array_almost_equal([v.y for v in surv_expected], [v.y for v in surv_actual])


def test_brier_scorer_no_predict_survival_function(make_whas500):
    with pytest.raises(
        AttributeError, match=r"FastSurvivalSVM\(\) object has no attribute 'predict_survival_function'"
    ):
        as_integrated_brier_score_scorer(FastSurvivalSVM(), times=[100, 200, 300])


@pytest.mark.parametrize("pred_func", ["predict_cumulative_hazard_function", "predict_survival_function"])
def test_scorer_no_predict_function(make_whas500, pred_func):
    whas500_data = make_whas500(to_numeric=True)
    scorer = as_concordance_index_ipcw_scorer(FastSurvivalSVM())
    scorer.fit(whas500_data.x, whas500_data.y)

    with pytest.raises(AttributeError, match=f"This 'as_concordance_index_ipcw_scorer' has no attribute {pred_func!r}"):
        getattr(scorer, pred_func)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numbers

import numpy as np
from scipy import stats
from sklearn.base import BaseEstimator
from sklearn.utils._param_validation import Interval, StrOptions
from sklearn.utils.validation import (
    check_array,
    check_consistent_length,
    check_is_fitted,
)

from .util import check_y_survival

__all__ = [
    "CensoringDistributionEstimator",
    "kaplan_meier_estimator",
    "nelson_aalen_estimator",
    "ipc_weights",
    "SurvivalFunctionEstimator",
]


def _compute_counts(event, time, order=None):
    """Count right censored and uncensored samples at each unique time point.

    Parameters
    ----------
    event : array
        Boolean event indicator.

    time : array
        Survival time or time of censoring.

    order : array or None
        Indices to order time in ascending order.
        If None, order will be computed.

    Returns
    -------
    times : array
        Unique time points.

    n_events : array
        Number of events at each time point.

    n_at_risk : array
        Number of samples that have not been censored or have not had an event at each time point.

    n_censored : array
        Number of censored samples at each time point.
    """
    n_samples = event.shape[0]

    if order is None:
        order = np.argsort(time, kind="mergesort")

    uniq_times = np.empty(n_samples, dtype=time.dtype)
    uniq_events = np.empty(n_samples, dtype=int)
    uniq_counts = np.empty(n_samples, dtype=int)

    i = 0
    prev_val = time[order[0]]
    j = 0
    while True:
        count_event = 0
        count = 0
        while i < n_samples and prev_val == time[order[i]]:
            if event[order[i]]:
                count_event += 1

            count += 1
            i += 1

        uniq_times[j] = prev_val
        uniq_events[j] = count_event
        uniq_counts[j] = count
        j += 1

        if i == n_samples:
            break

        prev_val = time[order[i]]

    times = np.resize(uniq_times, j)
    n_events = np.resize(uniq_events, j)
    total_count = np.resize(uniq_counts, j)
    n_censored = total_count - n_events

    # offset cumulative sum by one
    total_count = np.r_[0, total_count]
    n_at_risk = n_samples - np.cumsum(total_count)

    return times, n_events, n_at_risk[:-1], n_censored


def _compute_counts_truncated(event, time_enter, time_exit):
    """Compute counts for left truncated and right censored survival data.

    Parameters
    ----------
    event : array
        Boolean event indicator.

    time_start : array
        Time when a subject entered the study.

    time_exit : array
        Time when a subject left the study due to an
        event or censoring.

    Returns
    -------
    times : array
        Unique time points.

    n_events : array
        Number of events at each time point.

    n_at_risk : array
        Number of samples that are censored or have an event at each time point.
    """
    if (time_enter > time_exit).any():
        raise ValueError("exit time must be larger start time for all samples")

    n_samples = event.shape[0]

    uniq_times = np.sort(
        np.unique(np.r_[time_enter, time_exit]), kind="mergesort")
    total_counts = np.empty(len(uniq_times), dtype=int)
    event_counts = np.empty(len(uniq_times), dtype=int)

    order_enter = np.argsort(time_enter, kind="mergesort")
    order_exit = np.argsort(time_exit, kind="mergesort")
    s_time_enter = time_enter[order_enter]
    s_time_exit = time_exit[order_exit]

    t0 = uniq_times[0]
    # everything larger is included
    idx_enter = np.searchsorted(s_time_enter, t0, side="right")
    # everything smaller is excluded
    idx_exit = np.searchsorted(s_time_exit, t0, side="left")

    total_counts[0] = idx_enter
    # except people die on the day they enter
    event_counts[0] = 0

    for i in range(1, len(uniq_times)):
        ti = uniq_times[i]

        while idx_enter < n_samples and s_time_enter[idx_enter] < ti:
            idx_enter += 1

        while idx_exit < n_samples and s_time_exit[idx_exit] < ti:
            idx_exit += 1

        risk_set = np.setdiff1d(
            order_enter[:idx_enter], order_exit[:idx_exit], assume_unique=True)
        total_counts[i] = len(risk_set)

        count_event = 0
        k = idx_exit
        while k < n_samples and s_time_exit[k] == ti:
            if event[order_exit[k]]:
                count_event += 1
            k += 1
        event_counts[i] = count_event

    return uniq_times, event_counts, total_counts


def _ci_logmlog(prob_survival, sigma_t, z):
    """Compute the pointwise log-minus-log transformed confidence intervals"""
    eps = np.finfo(prob_survival.dtype).eps
    log_p = np.zeros_like(prob_survival)
    np.log(prob_survival, where=prob_survival > eps, out=log_p)
    theta = np.zeros_like(prob_survival)
    np.true_divide(sigma_t, log_p, where=log_p < -eps, out=theta)
    theta = np.array([[-1], [1]]) * theta * z
    ci = np.exp(np.exp(theta) * log_p)
    ci[:, prob_survival <= eps] = 0.0
    ci[:, 1.0 - prob_survival <= eps] = 1.0
    return ci


def _km_ci_estimator(prob_survival, ratio_var, conf_level, conf_type):
    if conf_type not in {"log-log"}:
        raise ValueError(
            f"conf_type must be None or a str among {{'log-log'}}, but was {conf_type!r}")

    if not isinstance(conf_level, numbers.Real) or not np.isfinite(conf_level) or conf_level <= 0 or conf_level >= 1.0:
        raise ValueError(
            f"conf_level must be a float in the range (0.0, 1.0), but was {conf_level!r}")

    z = stats.norm.isf((1.0 - conf_level) / 2.0)
    sigma = np.sqrt(np.cumsum(ratio_var))
    ci = _ci_logmlog(prob_survival, sigma, z)
    return ci


def kaplan_meier_estimator(
    event,
    time_exit,
    time_enter=None,
    time_min=None,
    reverse=False,
    conf_level=0.95,
    conf_type=None,
):
    """Kaplan-Meier estimator of survival function.

    See [1]_ for further description.

    Parameters
    ----------
    event : array-like, shape = (n_samples,)
        Contains binary event indicators.

    time_exit : array-like, shape = (n_samples,)
        Contains event/censoring times.

    time_enter : array-like, shape = (n_samples,), optional
        Contains time when each individual entered the study for
        left truncated survival data.

    time_min : float, optional
        Compute estimator conditional on survival at least up to
        the specified time.

    reverse : bool, optional, default: False
        Whether to estimate the censoring distribution.
        When there are ties between times at which events are observed,
        then events come first and are subtracted from the denominator.
        Only available for right-censored data, i.e. `time_enter` must
        be None.

    conf_level : float, optional, default: 0.95
        The level for a two-sided confidence interval on the survival curves.

    conf_type : None or {'log-log'}, optional, default: None.
        The type of confidence intervals to estimate.
        If `None`, no confidence intervals are estimated.
        If "log-log", estimate confidence intervals using
        the log hazard or :math:`log(-log(S(t)))` as described in [2]_.

    Returns
    -------
    time : array, shape = (n_times,)
        Unique times.

    prob_survival : array, shape = (n_times,)
        Survival probability at each unique time point.
        If `time_enter` is provided, estimates are conditional probabilities.

    conf_int : array, shape = (2, n_times)
        Pointwise confidence interval of the Kaplan-Meier estimator
        at each unique time point.
        Only provided if `conf_type` is not None.

    Examples
    --------
    Creating a Kaplan-Meier curve:

    >>> x, y, conf_int = kaplan_meier_estimator(event, time, conf_type="log-log")
    >>> plt.step(x, y, where="post")
    >>> plt.fill_between(x, conf_int[0], conf_int[1], alpha=0.25, step="post")
    >>> plt.ylim(0, 1)
    >>> plt.show()

    See also
    --------
    survivalist.nonparametric.SurvivalFunctionEstimator
        Estimator API of the Kaplan-Meier estimator.

    References
    ----------
    .. [1] Kaplan, E. L. and Meier, P., "Nonparametric estimation from incomplete observations",
           Journal of The American Statistical Association, vol. 53, pp. 457-481, 1958.
    .. [2] Borgan Ø. and Liestøl K., "A Note on Confidence Intervals and Bands for the
           Survival Function Based on Transformations", Scandinavian Journal of
           Statistics. 1990;17(1):35–41.
    """
    event, time_enter, time_exit = check_y_survival(
        event, time_enter, time_exit, allow_all_censored=True)
    check_consistent_length(event, time_enter, time_exit)

    if conf_type is not None and reverse:
        raise NotImplementedError(
            "Confidence intervals of the censoring distribution is not implemented.")

    if time_enter is None:
        uniq_times, n_events, n_at_risk, n_censored = _compute_counts(
            event, time_exit)

        if reverse:
            n_at_risk -= n_events
            n_events = n_censored
    else:
        if reverse:
            raise ValueError(
                "The censoring distribution cannot be estimated from left truncated data")

        uniq_times, n_events, n_at_risk = _compute_counts_truncated(
            event, time_enter, time_exit)

    # account for 0/0 = nan
    ratio = np.divide(
        n_events,
        n_at_risk,
        out=np.zeros(uniq_times.shape[0], dtype=float),
        where=n_events != 0,
    )
    values = 1.0 - ratio

    if conf_type is not None:
        ratio_var = np.divide(
            n_events,
            n_at_risk * (n_at_risk - n_events),
            out=np.zeros(uniq_times.shape[0], dtype=float),
            where=(n_events != 0) & (n_at_risk != n_events),
        )

    if time_min is not None:
        mask = uniq_times >= time_min
        uniq_times = np.compress(mask, uniq_times)
        values = np.compress(mask, values)

    prob_survival = np.cumprod(values)

    if conf_type is None:
        return uniq_times, prob_survival

    if time_min is not None:
        ratio_var = np.compress(mask, ratio_var)

    ci = _km_ci_estimator(prob_survival, ratio_var, conf_level, conf_type)

    return uniq_times, prob_survival, ci


def nelson_aalen_estimator(event, time):
    """Nelson-Aalen estimator of cumulative hazard function.

    See [1]_, [2]_ for further description.

    Parameters
    ----------
    event : array-like, shape = (n_samples,)
        Contains binary event indicators.

    time : array-like, shape = (n_samples,)
        Contains event/censoring times.

    Returns
    -------
    time : array, shape = (n_times,)
        Unique times.

    cum_hazard : array, shape = (n_times,)
        Cumulative hazard at each unique time point.

    References
    ----------
    .. [1] Nelson, W., "Theory and applications of hazard plotting for censored failure data",
           Technometrics, vol. 14, pp. 945-965, 1972.

    .. [2] Aalen, O. O., "Nonparametric inference for a family of counting processes",
           Annals of Statistics, vol. 6, pp. 701–726, 1978.
    """
    event, time = check_y_survival(event, time)
    check_consistent_length(event, time)
    uniq_times, n_events, n_at_risk, _ = _compute_counts(event, time)

    y = np.cumsum(n_events / n_at_risk)

    return uniq_times, y


def ipc_weights(event, time):
    """Compute inverse probability of censoring weights

    Parameters
    ----------
    event : array, shape = (n_samples,)
        Boolean event indicator.

    time : array, shape = (n_samples,)
        Time when a subject experienced an event or was censored.

    Returns
    -------
    weights : array, shape = (n_samples,)
        inverse probability of censoring weights

    See also
    --------
    CensoringDistributionEstimator
        An estimator interface for estimating inverse probability
        of censoring weights for unseen time points.
    """
    if event.all():
        return np.ones(time.shape[0])

    unique_time, p = kaplan_meier_estimator(event, time, reverse=True)

    idx = np.searchsorted(unique_time, time[event])
    Ghat = p[idx]

    assert (Ghat > 0).all()

    weights = np.zeros(time.shape[0])
    weights[event] = 1.0 / Ghat

    return weights


class SurvivalFunctionEstimator(BaseEstimator):
    """Kaplan–Meier estimate of the survival function.

    Parameters
    ----------
    conf_level : float, optional, default: 0.95
        The level for a two-sided confidence interval on the survival curves.

    conf_type : None or {'log-log'}, optional, default: None.
        The type of confidence intervals to estimate.
        If `None`, no confidence intervals are estimated.
        If "log-log", estimate confidence intervals using
        the log hazard or :math:`log(-log(S(t)))`.

    See also
    --------
    survivalist.nonparametric.kaplan_meier_estimator
        Functional API of the Kaplan-Meier estimator.
    """

    _parameter_constraints = {
        "conf_level": [Interval(numbers.Real, 0.0, 1.0, closed="neither")],
        "conf_type": [None, StrOptions({"log-log"})],
    }

    def __init__(self, conf_level=0.95, conf_type=None):
        self.conf_level = conf_level
        self.conf_type = conf_type

    def fit(self, y):
        """Estimate survival distribution from training data.

        Parameters
        ----------
        y : structured array, shape = (n_samples,)
            A structured array containing the binary event indicator
            as first field, and time of event or time of censoring as
            second field.

        Returns
        -------
        self
        """
        self._validate_params()
        event, time = check_y_survival(y, allow_all_censored=True)

        values = kaplan_meier_estimator(
            event, time, conf_level=self.conf_level, conf_type=self.conf_type)
        if self.conf_type is None:
            unique_time, prob = values
        else:
            unique_time, prob, conf_int = values
            self.conf_int_ = np.column_stack((np.ones((2, 1)), conf_int))

        self.unique_time_ = np.r_[-np.inf, unique_time]
        self.prob_ = np.r_[1.0, prob]

        return self

    def predict_proba(self, time, return_conf_int=False):
        """Return probability of an event after given time point.

        :math:`\\hat{S}(t) = P(T > t)`

        Parameters
        ----------
        time : array, shape = (n_samples,)
            Time to estimate probability at.

        return_conf_int : bool, optional, default: False
            Whether to return the pointwise confidence interval
            of the survival function.
            Only available if :meth:`fit()` has been called
            with the `conf_type` parameter set.

        Returns
        -------
        prob : array, shape = (n_samples,)
            Probability of an event at the passed time points.

        conf_int : array, shape = (2, n_samples)
            Pointwise confidence interval at the passed time points.
            Only provided if `return_conf_int` is True.
        """
        check_is_fitted(self, "unique_time_")
        if return_conf_int and not hasattr(self, "conf_int_"):
            raise ValueError(
                "If return_conf_int is True, SurvivalFunctionEstimator must be fitted with conf_int != None"
            )

        time = check_array(time, ensure_2d=False,
                           estimator=self, input_name="time")

        # K-M is undefined if estimate at last time point is non-zero
        extends = time > self.unique_time_[-1]
        if self.prob_[-1] > 0 and extends.any():
            raise ValueError(
                f"time must be smaller than largest observed time point: {self.unique_time_[-1]}")

        # beyond last time point is zero probability
        Shat = np.empty(time.shape, dtype=float)
        Shat[extends] = 0.0

        valid = ~extends
        time = time[valid]
        idx = np.searchsorted(self.unique_time_, time)
        # for non-exact matches, we need to shift the index to left
        eps = np.finfo(self.unique_time_.dtype).eps
        exact = np.absolute(self.unique_time_[idx] - time) < eps
        idx[~exact] -= 1
        Shat[valid] = self.prob_[idx]

        if not return_conf_int:
            return Shat

        ci = np.empty((2, time.shape[0]), dtype=float)
        ci[:, extends] = np.nan
        ci[:, valid] = self.conf_int_[:, idx]
        return Shat, ci


class CensoringDistributionEstimator(SurvivalFunctionEstimator):
    """Kaplan–Meier estimator for the censoring distribution."""

    def fit(self, y):
        """Estimate censoring distribution from training data.

        Parameters
        ----------
        y : structured array, shape = (n_samples,)
            A structured array containing the binary event indicator
            as first field, and time of event or time of censoring as
            second field.

        Returns
        -------
        self
        """
        event, time = check_y_survival(y)
        if event.all():
            self.unique_time_ = np.unique(time)
            self.prob_ = np.ones(self.unique_time_.shape[0])
        else:
            unique_time, prob = kaplan_meier_estimator(
                event, time, reverse=True)
            self.unique_time_ = np.r_[-np.inf, unique_time]
            self.prob_ = np.r_[1.0, prob]

        return self

    def predict_ipcw(self, y):
        """Return inverse probability of censoring weights at given time points.

        :math:`\\omega_i = \\delta_i / \\hat{G}(y_i)`

        Parameters
        ----------
        y : structured array, shape = (n_samples,)
            A structured array containing the binary event indicator
            as first field, and time of event or time of censoring as
            second field.

        Returns
        -------
        ipcw : array, shape = (n_samples,)
            Inverse probability of censoring weights.
        """
        event, time = check_y_survival(y)
        Ghat = self.predict_proba(time[event])

        if (Ghat == 0.0).any():
            raise ValueError(
                "censoring survival function is zero at one or more time points")

        weights = np.zeros(time.shape[0])
        weights[event] = 1.0 / Ghat

        return weights



import os
from pathlib import Path
import shutil
import sys

from packaging import version
from setuptools import Command, Extension, setup

CYTHON_MIN_VERSION = version.parse("0.29")


# adapted from bottleneck's setup.py
class clean(Command):
    user_options = [("all", "a", "")]

    def initialize_options(self):
        self.all = True
        self.delete_dirs = []
        self.delete_files = []

        for root, dirs, files in os.walk("survivalist"):
            root = Path(root)
            for d in dirs:
                if d == "__pycache__":
                    self.delete_dirs.append(root / d)

            if "__pycache__" in root.name:
                continue

            for f in (root / x for x in files):
                ext = f.suffix
                if ext == ".pyc" or ext == ".so":
                    self.delete_files.append(f)

                if ext in (
                    ".c",
                    ".cpp",
                ):
                    source_file = f.with_suffix(".pyx")
                    if source_file.exists():
                        self.delete_files.append(f)

        build_path = Path("build")
        if build_path.exists():
            self.delete_dirs.append(build_path)

    def finalize_options(self):
        pass

    def run(self):
        for delete_dir in self.delete_dirs:
            shutil.rmtree(delete_dir)
        for delete_file in self.delete_files:
            delete_file.unlink()


EXTENSIONS = {
    "_coxph_loss": {"sources": ["survivalist/ensemble/_coxph_loss.pyx"]},
}


def get_module_from_sources(sources):
    for src_path in map(Path, sources):
        if src_path.suffix == ".pyx":
            return ".".join(src_path.parts[:-1] + (src_path.stem,))
    raise ValueError(f"could not find module from sources: {sources!r}")


def _check_cython_version():
    message = (
        f"Please install Cython with a version >= {CYTHON_MIN_VERSION} in order to build a scikit-learn from source."
    )
    try:
        import Cython
    except ModuleNotFoundError:
        # Re-raise with more informative error message instead:
        raise ModuleNotFoundError(message)

    if version.parse(Cython.__version__) < CYTHON_MIN_VERSION:
        message += f" The current version of Cython is {Cython.__version__} installed in {Cython.__path__}."
        raise ValueError(message)


def cythonize_extensions(extensions):
    """Check that a recent Cython is available and cythonize extensions"""
    _check_cython_version()
    from Cython.Build import cythonize

    # http://docs.cython.org/en/latest/src/userguide/source_files_and_compilation.html#cythonize-arguments
    directives = {"language_level": "3"}
    cy_cov = os.environ.get("CYTHON_COVERAGE", False)
    macros = [("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]
    if cy_cov:
        directives["linetrace"] = True
        macros.append(("CYTHON_TRACE", "1"))
        macros.append(("CYTHON_TRACE_NOGIL", "1"))

    for ext in extensions:
        if ext.define_macros is None:
            ext.define_macros = macros
        else:
            ext.define_macros += macros

    return cythonize(extensions, compiler_directives=directives)


def get_extensions():
    import numpy

    numpy_includes = [numpy.get_include()]
    
    extensions = []
    for config in EXTENSIONS.values():
        name = get_module_from_sources(config["sources"])
        include_dirs = numpy_includes + config.get("include_dirs", [])
        extra_compile_args = config.get("extra_compile_args", [])
        language = config.get("language", "c")
        ext = Extension(
            name=name,
            sources=config["sources"],
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args,
            language=language,
        )
        extensions.append(ext)

    # Skip cythonization as we do not want to include the generated
    # C/C++ files in the release tarballs as they are not necessarily
    # forward compatible with future versions of Python for instance.
    if "sdist" not in sys.argv and "clean" not in sys.argv:
        extensions = cythonize_extensions(extensions)

    return extensions


if __name__ == "__main__":
    setup(
        ext_modules=get_extensions(),
        cmdclass={"clean": clean},
        version="0.3.0"
    )



from collections import OrderedDict

import numpy as np
import pandas as pd
from scipy import stats
from sklearn.utils.validation import check_array

from .util import check_array_survival

__all__ = ["compare_survival"]


def compare_survival(y, group_indicator, return_stats=False):
    """K-sample log-rank hypothesis test of identical survival functions.

    Compares the pooled hazard rate with each group-specific
    hazard rate. The alternative hypothesis is that the hazard
    rate of at least one group differs from the others at some time.

    See [1]_ for more details.

    Parameters
    ----------
    y : structured array, shape = (n_samples,)
        A structured array containing the binary event indicator
        as first field, and time of event or time of censoring as
        second field.

    group_indicator : array-like, shape = (n_samples,)
        Group membership of each sample.

    return_stats : bool, optional, default: False
        Whether to return a data frame with statistics for each group
        and the covariance matrix of the test statistic.

    Returns
    -------
    chisq : float
        Test statistic.
    pvalue : float
        Two-sided p-value with respect to the null hypothesis
        that the hazard rates across all groups are equal.
    stats : pandas.DataFrame
        Summary statistics for each group:  number of samples,
        observed number of events, expected number of events,
        and test statistic.
        Only provided if `return_stats` is True.
    covariance : array, shape=(n_groups, n_groups)
        Covariance matrix of the test statistic.
        Only provided if `return_stats` is True.

    References
    ----------
    .. [1] Fleming, T. R. and Harrington, D. P.
           A Class of Hypothesis Tests for One and Two Samples of Censored Survival Data.
           Communications In Statistics 10 (1981): 763-794.
    """

    event, time = check_array_survival(group_indicator, y)
    group_indicator = check_array(
        group_indicator,
        dtype="O",
        ensure_2d=False,
        estimator="compare_survival",
        input_name="group_indicator",
    )

    n_samples = time.shape[0]
    groups, group_counts = np.unique(group_indicator, return_counts=True)
    n_groups = groups.shape[0]
    if n_groups == 1:
        raise ValueError(
            "At least two groups must be specified, but only one was provided.")

    # sort descending
    o = np.argsort(-time, kind="mergesort")
    x = group_indicator[o]
    event = event[o]
    time = time[o]

    at_risk = np.zeros(n_groups, dtype=int)
    observed = np.zeros(n_groups, dtype=int)
    expected = np.zeros(n_groups, dtype=float)
    covar = np.zeros((n_groups, n_groups), dtype=float)

    covar_indices = np.diag_indices(n_groups)

    k = 0
    while k < n_samples:
        ti = time[k]
        total_events = 0
        while k < n_samples and ti == time[k]:
            idx = np.searchsorted(groups, x[k])
            if event[k]:
                observed[idx] += 1
                total_events += 1
            at_risk[idx] += 1
            k += 1

        if total_events != 0:
            total_at_risk = k
            expected += at_risk * (total_events / total_at_risk)
            if total_at_risk > 1:
                multiplier = total_events * \
                    (total_at_risk - total_events) / \
                    (total_at_risk * (total_at_risk - 1))
                temp = at_risk * multiplier
                covar[covar_indices] += temp
                covar -= np.outer(temp, at_risk) / total_at_risk

    df = n_groups - 1
    zz = observed[:df] - expected[:df]
    chisq = np.linalg.solve(covar[:df, :df], zz).dot(zz)
    pval = stats.chi2.sf(chisq, df)

    if return_stats:
        table = OrderedDict()
        table["counts"] = group_counts
        table["observed"] = observed
        table["expected"] = expected
        table["statistic"] = observed - expected
        table = pd.DataFrame.from_dict(table)
        table.index = pd.Index(groups, name="group", dtype=groups.dtype)
        return chisq, pval, table, covar

    return chisq, pval



from .base import (
    get_x_y,  # noqa: F401
    load_aids,  # noqa: F401
    load_arff_files_standardized,  # noqa: F401
    load_breast_cancer,  # noqa: F401
    load_flchain,  # noqa: F401
    load_gbsg2,  # noqa: F401
    load_veterans_lung_cancer,  # noqa: F401
    load_whas500,  # noqa: F401
)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os.path
import re

import numpy as np
import pandas as pd
from pandas.api.types import CategoricalDtype, is_object_dtype

_ILLEGAL_CHARACTER_PAT = re.compile(r"[^-_=\w\d\(\)<>\.]")


def writearff(data, filename, relation_name=None, index=True):
    """Write ARFF file

    Parameters
    ----------
    data : :class:`pandas.DataFrame`
        DataFrame containing data

    filename : string or file-like object
        Path to ARFF file or file-like object. In the latter case,
        the handle is closed by calling this function.

    relation_name : string, optional, default: "pandas"
        Name of relation in ARFF file.

    index : boolean, optional, default: True
        Write row names (index)
    """
    if isinstance(filename, str):
        fp = open(filename, "w")

        if relation_name is None:
            relation_name = os.path.basename(filename)
    else:
        fp = filename

        if relation_name is None:
            relation_name = "pandas"

    try:
        data = _write_header(data, fp, relation_name, index)
        fp.write("\n")
        _write_data(data, fp)
    finally:
        fp.close()


def _write_header(data, fp, relation_name, index):
    """Write header containing attribute names and types"""
    fp.write(f"@relation {relation_name}\n\n")

    if index:
        data = data.reset_index()

    attribute_names = _sanitize_column_names(data)

    for column, series in data.items():
        name = attribute_names[column]
        fp.write(f"@attribute {name}\t")

        if isinstance(series.dtype, CategoricalDtype) or is_object_dtype(series):
            _write_attribute_categorical(series, fp)
        elif np.issubdtype(series.dtype, np.floating):
            fp.write("real")
        elif np.issubdtype(series.dtype, np.integer):
            fp.write("integer")
        elif np.issubdtype(series.dtype, np.datetime64):
            fp.write("date 'yyyy-MM-dd HH:mm:ss'")
        else:
            raise TypeError(f"unsupported type {series.dtype}")

        fp.write("\n")
    return data


def _sanitize_column_names(data):
    """Replace illegal characters with underscore"""
    new_names = {}
    for name in data.columns:
        new_names[name] = _ILLEGAL_CHARACTER_PAT.sub("_", name)
    return new_names


def _check_str_value(x):
    """If string has a space, wrap it in double quotes and remove/escape illegal characters"""
    if isinstance(x, str):
        # remove commas, and single quotation marks since loadarff cannot deal with it
        x = x.replace(",", ".").replace(
            chr(0x2018), "'").replace(chr(0x2019), "'")

        # put string in double quotes
        if " " in x:
            if x[0] in ('"', "'"):
                x = x[1:]
            if x[-1] in ('"', "'"):
                x = x[: len(x) - 1]
            x = '"' + x.replace('"', '\\"') + '"'
    return str(x)


_check_str_array = np.frompyfunc(_check_str_value, 1, 1)


def _write_attribute_categorical(series, fp):
    """Write categories of a categorical/nominal attribute"""
    if isinstance(series.dtype, CategoricalDtype):
        categories = series.cat.categories
        string_values = _check_str_array(categories)
    else:
        categories = series.dropna().unique()
        string_values = sorted(_check_str_array(
            categories), key=lambda x: x.strip('"'))

    values = ",".join(string_values)
    fp.write("{")
    fp.write(values)
    fp.write("}")


def _write_data(data, fp):
    """Write the data section"""
    fp.write("@data\n")

    def to_str(x):
        if pd.isnull(x):
            return "?"
        return str(x)

    data = data.applymap(to_str)
    n_rows = data.shape[0]
    for i in range(n_rows):
        str_values = list(data.iloc[i, :].apply(_check_str_array))
        line = ",".join(str_values)
        fp.write(line)
        fp.write("\n")



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from importlib import import_module
import inspect
from pathlib import Path
import pkgutil

import numpy as np
from numpy.testing import assert_almost_equal, assert_array_equal
import pytest
from sklearn.base import BaseEstimator, TransformerMixin

import survivalist
from survivalist.metrics import concordance_index_censored


def assert_cindex_almost_equal(event_indicator, event_time, estimate, expected):
    result = concordance_index_censored(event_indicator, event_time, estimate)
    assert_array_equal(result[1:], expected[1:])
    concordant, discordant, tied_risk = result[1:4]
    cc = (concordant + 0.5 * tied_risk) / (concordant + discordant + tied_risk)
    assert_almost_equal(result[0], cc)
    assert_almost_equal(result[0], expected[0])


def assert_survival_function_properties(surv_fns):
    if not np.isfinite(surv_fns).all():
        raise AssertionError(
            "survival function contains values that are not finite")
    if np.any(surv_fns < 0.0):
        raise AssertionError("survival function contains negative values")
    if np.any(surv_fns > 1.0):
        raise AssertionError("survival function contains values larger 1")

    d = np.apply_along_axis(np.diff, 1, surv_fns)
    if np.any(d > 0):
        raise AssertionError(
            "survival functions are not monotonically decreasing")

    # survival function at first time point
    num_closer_to_zero = np.sum(1.0 - surv_fns[:, 0] >= surv_fns[:, 0])
    if num_closer_to_zero / surv_fns.shape[0] > 0.5:
        raise AssertionError(
            f"most ({num_closer_to_zero}) probabilities at first time point are closer to 0 than 1")

    # survival function at last time point
    num_closer_to_one = np.sum(1.0 - surv_fns[:, -1] < surv_fns[:, -1])
    if num_closer_to_one / surv_fns.shape[0] > 0.5:
        raise AssertionError(
            f"most ({num_closer_to_one}) probabilities at last time point are closer to 1 than 0")


def assert_chf_properties(chf):
    if not np.isfinite(chf).all():
        raise AssertionError("chf contains values that are not finite")
    if np.any(chf < 0.0):
        raise AssertionError("chf contains negative values")

    d = np.apply_along_axis(np.diff, 1, chf)
    if np.any(d < 0):
        raise AssertionError("chf are not monotonically increasing")

    # chf at first time point
    num_closer_to_one = np.sum(1.0 - chf[:, 0] < chf[:, 0])
    if num_closer_to_one / chf.shape[0] > 0.5:
        raise AssertionError(
            f"most ({num_closer_to_one}) hazard rates at first time point are closer to 1 than 0")


def _is_survival_estimator(x):
    return (
        inspect.isclass(x)
        and issubclass(x, BaseEstimator)
        and not issubclass(x, TransformerMixin)
        and x.__module__.startswith("survivalist.")
        and not x.__name__.startswith("_")
        and x.__module__.split(".", 2)[1] not in {"metrics", "nonparametric"}
    )


def all_survival_estimators():
    root = str(Path(survivalist.__file__).parent)
    all_classes = []
    for _importer, modname, _ispkg in pkgutil.walk_packages(path=[root], prefix="survivalist."):
        # meta-estimators require base estimators
        if modname.startswith("survivalist.meta"):
            continue
        module = import_module(modname)
        for _name, cls in inspect.getmembers(module, _is_survival_estimator):
            if inspect.isabstract(cls):
                continue
            all_classes.append(cls)
    return set(all_classes)


class FixtureParameterFactory:
    def get_cases(self):
        cases = []
        for name, func in inspect.getmembers(self):
            if name.startswith("data_"):
                values = func()
                cases.append(pytest.param(*values, id=name))
        return cases



from contextlib import nullcontext as does_not_raise
import itertools
import math
from os.path import dirname, join

import numpy as np
from numpy.testing import assert_array_almost_equal, assert_array_equal
import pandas as pd
import pytest
from sklearn.metrics import mean_squared_error

from survivalist.column import categorical_to_numeric, standardize
from survivalist.datasets import load_whas500
from survivalist.ensemble import ComponentwiseGradientBoostingSurvivalAnalysis, GradientBoostingSurvivalAnalysis
from survivalist.testing import FixtureParameterFactory, assert_cindex_almost_equal
from survivalist.util import Surv

CGBOOST_CUMHAZ_FILE = join(dirname(__file__), "data", "compnentwise-gradient-boosting-coxph-cumhazard.csv")
CGBOOST_SURV_FILE = join(dirname(__file__), "data", "compnentwise-gradient-boosting-coxph-surv.csv")
GBOOST_CUMHAZ_FILE = join(dirname(__file__), "data", "gradient-boosting-coxph-cumhazard.csv")
GBOOST_SURV_FILE = join(dirname(__file__), "data", "gradient-boosting-coxph-surv.csv")


def early_stopping_monitor(i, est, locals_):  # pylint: disable=unused-argument
    """Returns True on the 10th iteration."""
    return i == 9


class MaxFeaturesCases(FixtureParameterFactory):
    _prefix = (
        "The 'max_features' parameter of GradientBoostingSurvivalAnalysis must be "
        r"an int in the range \[1, inf\), a float in the range \(0\.0, 1\.0\], a str "
        r"among {.+} or None\. "
    )

    @property
    def n_features(self):
        return 14

    def data_1p0(self):
        return 1.0, self.n_features, does_not_raise()

    def data_sqrt(self):
        return "sqrt", np.sqrt(self.n_features), does_not_raise()

    def data_log2(self):
        return "log2", np.log2(self.n_features), does_not_raise()

    def data_0p25(self):
        return 0.25, 0.25 * self.n_features, does_not_raise()

    def data_5(self):
        return 5, 5, does_not_raise()

    def data_m1(self):
        return -1, None, pytest.raises(ValueError, match=MaxFeaturesCases._prefix + r"Got -1 instead\.")

    def data_m1p125(self):
        return -1.125, None, pytest.raises(ValueError, match=MaxFeaturesCases._prefix + r"Got -1\.125 instead\.")

    def data_fail_me(self):
        return "fail_me", None, pytest.raises(ValueError, match=MaxFeaturesCases._prefix + r"Got 'fail_me' instead\.")


class TestGradientBoosting:
    @property
    def data(self):
        x, y = load_whas500()
        x = categorical_to_numeric(standardize(x, with_std=False))
        return x.values, y

    def assert_fit_and_predict(self, expected_cindex, mask_test=None, **kwargs):
        X, y = self.data
        if mask_test is None:
            X_train, y_train, X_test, y_test = X, y, X, y
        else:
            X_train = X[~mask_test]
            y_train = y[~mask_test]
            X_test = X[mask_test]
            y_test = y[mask_test]

        kwargs.setdefault("random_state", 0)
        model = GradientBoostingSurvivalAnalysis(**kwargs)
        model.fit(X_train, y_train)

        if expected_cindex is not None:
            p = model.predict(X_test)

            assert_cindex_almost_equal(
                y_test["fstat"],
                y_test["lenfol"],
                p,
                expected_cindex,
            )

        return model

    def test_fit(self):
        model = self.assert_fit_and_predict(
            expected_cindex=(0.86272605091218779, 64826, 10309, 14, 14),
            n_estimators=100,
            max_depth=3,
            min_samples_split=10,
        )

        assert model.max_features_ == 14
        assert not hasattr(model, "oob_improvement_")

        assert (100,) == model.train_score_.shape

    def test_predict_incorrect_features(self):
        model = self.assert_fit_and_predict(expected_cindex=None)
        with pytest.raises(
            ValueError,
            match="X has 2 features, but GradientBoostingSurvivalAnalysis is expecting 14 features as input.",
        ):
            model.predict(np.random.randn(10, 2))

    def test_fit_subsample(self):
        idx = np.array(
            [
                229,
                337,
                327,
                416,
                306,
                131,
                5,
                431,
                434,
                134,
                276,
                264,
                34,
                274,
                489,
                11,
                378,
                467,
                425,
                241,
                93,
                310,
                189,
                277,
                138,
                332,
                313,
                59,
                212,
                171,
                375,
                85,
                266,
                36,
                272,
                287,
                185,
                338,
                33,
                343,
                363,
                9,
                204,
                472,
                48,
                280,
                15,
                184,
                24,
                82,
            ],
            dtype=int,
        )
        incl_mask = np.zeros(500, dtype=bool)
        incl_mask[idx] = True

        model = self.assert_fit_and_predict(
            expected_cindex=(0.7754172989377845, 511, 148, 0, 0),
            mask_test=incl_mask,
            n_estimators=50,
            learning_rate=0.2,
            min_samples_split=10,
            max_features=1.0,
            subsample=0.637,
        )

        assert model.max_features_ == 14
        assert hasattr(model, "oob_improvement_")

        assert (50,) == model.train_score_.shape
        assert (50,) == model.oob_improvement_.shape

    @pytest.mark.slow()
    def test_fit_dropout(self):
        model = self.assert_fit_and_predict(
            expected_cindex=(0.9051950, 68024, 7124, 1, 14),
            n_estimators=100,
            max_features=8,
            learning_rate=1.0,
            dropout_rate=0.03,
        )

        assert not hasattr(model, "oob_improvement_")
        assert model.max_features_ == 8

    @pytest.mark.parametrize(
        "parameter,value",
        [
            ("max_depth", 4.0),
            ("n_estimators", 100.0),
            ("max_leaf_nodes", 15.0),
            ("min_samples_split", 10.0),
        ],
    )
    def test_fit_int_param_as_float(self, parameter, value):
        kwargs = {parameter: value}
        model = GradientBoostingSurvivalAnalysis(**kwargs)

        msg = f"The '{parameter}' parameter of GradientBoostingSurvivalAnalysis must be an int in the range "

        X, y = self.data
        with pytest.raises(ValueError, match=msg):
            model.fit(X, y)

    @pytest.mark.parametrize(
        "fn,expected_file",
        [("predict_survival_function", GBOOST_SURV_FILE), ("predict_cumulative_hazard_function", GBOOST_CUMHAZ_FILE)],
    )
    def test_predict_function(self, fn, expected_file):
        X, y = self.data

        train_y = y[10:]
        test_x = X[:10]
        test_mask = np.zeros(500, dtype=bool)
        test_mask[:10] = True

        model = self.assert_fit_and_predict(
            expected_cindex=None,
            mask_test=test_mask,
            n_estimators=100,
            max_depth=2,
        )

        surv_fn = getattr(model, fn)(test_x)

        times = np.unique(train_y["lenfol"][train_y["fstat"]])
        actual = np.vstack([fn_gb(times) for fn_gb in surv_fn])

        expected = np.loadtxt(expected_file, delimiter=",")

        assert_array_almost_equal(actual, expected)

    @pytest.mark.parametrize("max_features,expected_features,expected_error", MaxFeaturesCases().get_cases())
    def test_max_features(self, max_features, expected_features, expected_error):
        X, y = self.data

        model = GradientBoostingSurvivalAnalysis(
            n_estimators=10,
            max_features=max_features,
            max_depth=3,
            random_state=0,
        )

        with expected_error:
            model.fit(X, y)

        if expected_features is not None:
            assert model.max_features_ == int(expected_features)

    def test_ccp_alpha(self):
        est_full = self.assert_fit_and_predict(
            expected_cindex=None,
            n_estimators=10,
            max_leaf_nodes=20,
            random_state=1,
        )

        est_pruned = self.assert_fit_and_predict(
            expected_cindex=None,
            n_estimators=10,
            max_leaf_nodes=20,
            ccp_alpha=10.0,
            random_state=1,
        )

        tree = est_full.estimators_[0, 0].tree_
        subtree = est_pruned.estimators_[0, 0].tree_
        assert tree.node_count > subtree.node_count
        assert tree.max_depth > subtree.max_depth

    def test_early_stopping(self):
        X, y = self.data

        model = GradientBoostingSurvivalAnalysis(
            n_estimators=1000,
            max_depth=2,
            n_iter_no_change=3,
            validation_fraction=0.2,
            random_state=0,
        )
        model.fit(X, y)

        assert model.n_estimators_ == 36

    def test_early_stopping_with_dropout(self):
        X, y = self.data

        model = GradientBoostingSurvivalAnalysis(
            n_estimators=1000,
            max_depth=2,
            n_iter_no_change=3,
            dropout_rate=0.25,
            validation_fraction=0.2,
            random_state=0,
        )
        model.fit(X, y)

        assert model.n_estimators_ == 21

    @staticmethod
    def test_negative_ccp_alpha(make_whas500):
        whas500_data = make_whas500(with_std=False, to_numeric=True)

        clf = GradientBoostingSurvivalAnalysis()
        msg = (
            "The 'ccp_alpha' parameter of GradientBoostingSurvivalAnalysis must be a float in the range "
            r"\[0\.0, inf\)\. Got -1\.0 instead\."
        )

        clf.set_params(ccp_alpha=-1.0)
        with pytest.raises(ValueError, match=msg):
            clf.fit(whas500_data.x, whas500_data.y)

    @staticmethod
    def test_negative_n_iter_no_change(make_whas500):
        whas500_data = make_whas500(with_std=False, to_numeric=True)

        clf = GradientBoostingSurvivalAnalysis()
        msg = (
            "The 'n_iter_no_change' parameter of GradientBoostingSurvivalAnalysis must be an int in the range "
            r"\[1, inf\) or None\. Got -1 instead\."
        )

        clf.set_params(n_iter_no_change=-1)
        with pytest.raises(ValueError, match=msg):
            clf.fit(whas500_data.x, whas500_data.y)

    def test_fit_verbose(self):
        self.assert_fit_and_predict(expected_cindex=None, n_estimators=10, verbose=1)

    def test_ipcwls_loss(self):
        model = self.assert_fit_and_predict(
            expected_cindex=None,
            loss="ipcwls",
            n_estimators=100,
            max_depth=3,
        )

        X, y = self.data
        time_predicted = model.predict(X)
        time_true = y["lenfol"]
        event_true = y["fstat"]

        rmse_all = np.sqrt(mean_squared_error(time_true, time_predicted))
        assert rmse_all == pytest.approx(590.5441693629117, 1e-7)

        rmse_uncensored = np.sqrt(mean_squared_error(time_true[event_true], time_predicted[event_true]))
        assert rmse_uncensored == pytest.approx(392.97741487479743, 1e-7)

        cindex = model.score(X, y)
        assert cindex == pytest.approx(0.8979161399, 1e-7)

        with pytest.raises(ValueError, match="`fit` must be called with the loss option set to 'coxph'"):
            model.predict_survival_function(X)

        with pytest.raises(ValueError, match="`fit` must be called with the loss option set to 'coxph'"):
            model.predict_cumulative_hazard_function(X)

    def test_squared_loss(self):
        model = self.assert_fit_and_predict(
            expected_cindex=None,
            loss="squared",
            n_estimators=100,
            max_depth=3,
        )

        X, y = self.data
        time_predicted = model.predict(X)
        time_true = y["lenfol"]
        event_true = y["fstat"]

        rmse_all = np.sqrt(mean_squared_error(time_true, time_predicted))
        assert rmse_all == pytest.approx(580.23345259002951, 1e-7)

        rmse_uncensored = np.sqrt(mean_squared_error(time_true[event_true], time_predicted[event_true]))
        assert rmse_uncensored == pytest.approx(383.10639243317951, 1e-7)

        cindex = model.score(X, y)
        assert cindex == pytest.approx(0.9021810004, 1e-7)

        with pytest.raises(ValueError, match="`fit` must be called with the loss option set to 'coxph'"):
            model.predict_survival_function(X)

        with pytest.raises(ValueError, match="`fit` must be called with the loss option set to 'coxph'"):
            model.predict_cumulative_hazard_function(X)

    def assert_staged_predict(self, model, X):
        """test if prediction for last stage equals ``predict``"""
        y_pred = model.predict(X)

        # test if prediction for last stage equals ``predict``
        y_staged = None
        for y_staged in model.staged_predict(X):
            assert y_staged.shape == y_pred.shape

        assert_array_equal(y_pred, y_staged)

    @pytest.mark.parametrize("dropout_rate", [0.0, 0.03])
    @pytest.mark.parametrize("loss", ["ipcwls", "squared"])
    def test_loss_staged_predict(self, loss, dropout_rate):
        # Test whether staged decision function eventually gives
        # the same prediction.
        model = self.assert_fit_and_predict(
            expected_cindex=None,
            loss=loss,
            dropout_rate=dropout_rate,
            n_estimators=100,
            max_depth=3,
        )

        X, _ = self.data
        self.assert_staged_predict(model, X)

    def test_monitor_early_stopping(self):
        X, y = self.data

        est = GradientBoostingSurvivalAnalysis(
            loss="ipcwls",
            n_estimators=50,
            max_depth=1,
            subsample=0.5,
            random_state=0,
        )
        est.fit(X, y, monitor=early_stopping_monitor)

        assert est.n_estimators == 50  # this is not altered
        assert est.estimators_.shape[0] == 10
        assert est.train_score_.shape[0] == 10
        assert est.oob_improvement_.shape[0] == 10


class TestSparseGradientBoosting:
    def assert_fit_and_predict(self, data, **kwargs):
        model = GradientBoostingSurvivalAnalysis(random_state=0, **kwargs)
        model.fit(data.x_sparse, data.y)

        assert model.estimators_.shape[0] == 100
        assert model.train_score_.shape == (100,)

        sparse_predict = model.predict(data.x_sparse)
        dense_predict = model.predict(data.x_dense.values)

        assert_array_almost_equal(sparse_predict, dense_predict)

        return model

    @pytest.mark.parametrize("loss", ["coxph", "squared", "ipcwls"])
    def test_fit(self, whas500_sparse_data, loss):
        model = self.assert_fit_and_predict(
            whas500_sparse_data,
            loss=loss,
            n_estimators=100,
            max_depth=1,
            min_samples_split=10,
            subsample=0.5,
        )
        assert model.oob_improvement_.shape == (100,)

    @pytest.mark.parametrize("loss", ["coxph", "squared", "ipcwls"])
    @pytest.mark.slow()
    def test_dropout(self, whas500_sparse_data, loss):
        self.assert_fit_and_predict(
            whas500_sparse_data,
            loss=loss,
            n_estimators=100,
            max_depth=1,
            min_samples_split=10,
            dropout_rate=0.03,
        )


class TestComponentwiseGradientBoosting:
    @staticmethod
    def test_fit(make_whas500):
        whas500_data = make_whas500(with_std=False, to_numeric=True)

        model = ComponentwiseGradientBoostingSurvivalAnalysis(n_estimators=100)
        model.fit(whas500_data.x, whas500_data.y)

        assert model.loss == "coxph"

        p = model.predict(whas500_data.x)

        assert_cindex_almost_equal(
            whas500_data.y["fstat"], whas500_data.y["lenfol"], p, (0.7755659, 58283, 16866, 0, 14)
        )

        expected_coef = pd.Series(np.zeros(15, dtype=float), index=whas500_data.names)
        expected_coef["age"] = 0.040919
        expected_coef["hr"] = 0.004977
        expected_coef["diasbp"] = -0.003407
        expected_coef["bmi"] = -0.017938
        expected_coef["sho"] = 0.429904
        expected_coef["chf"] = 0.508211

        assert_array_almost_equal(expected_coef.values, model.coef_)

        assert (100,) == model.train_score_.shape

        with pytest.raises(
            ValueError,
            match="X has 2 features, but ComponentwiseGradientBoostingSurvivalAnalysis is "
            "expecting 14 features as input.",
        ):
            model.predict(whas500_data.x[:, :2])

    @staticmethod
    def test_fit_subsample(make_whas500):
        whas500_data = make_whas500(with_std=False, to_numeric=True)

        model = ComponentwiseGradientBoostingSurvivalAnalysis(n_estimators=100, subsample=0.6, random_state=0)
        model.fit(whas500_data.x, whas500_data.y)
        p = model.predict(whas500_data.x)

        assert_cindex_almost_equal(
            whas500_data.y["fstat"], whas500_data.y["lenfol"], p, (0.7750602, 58245, 16904, 0, 14)
        )

        expected_coef = pd.Series(np.zeros(15, dtype=float), index=whas500_data.names)
        expected_coef["age"] = 0.041299
        expected_coef["hr"] = 0.00487
        expected_coef["diasbp"] = -0.003381
        expected_coef["bmi"] = -0.017018
        expected_coef["sho"] = 0.433685
        expected_coef["chf"] = 0.510277

        assert_array_almost_equal(expected_coef.values, model.coef_)

        assert (100,) == model.train_score_.shape
        assert (100,) == model.oob_improvement_.shape

        with pytest.raises(
            ValueError,
            match="X has 2 features, but ComponentwiseGradientBoostingSurvivalAnalysis is "
            "expecting 14 features as input.",
        ):
            model.predict(whas500_data.x[:, :2])

    @staticmethod
    def test_fit_dropout(make_whas500):
        whas500_data = make_whas500(with_std=False, to_numeric=True)

        model = ComponentwiseGradientBoostingSurvivalAnalysis(
            n_estimators=100, learning_rate=1.0, dropout_rate=0.03, random_state=0
        )
        model.fit(whas500_data.x, whas500_data.y)
        p = model.predict(whas500_data.x)

        assert_cindex_almost_equal(
            whas500_data.y["fstat"], whas500_data.y["lenfol"], p, (0.7777482, 58447, 16702, 0, 14)
        )

        expected_coef = pd.Series(np.zeros(15, dtype=float), index=whas500_data.names)
        expected_coef["age"] = 0.248771
        expected_coef["hr"] = 0.04392
        expected_coef["diasbp"] = -0.022432
        expected_coef["bmi"] = -0.138095
        expected_coef["sho"] = 2.863096
        expected_coef["chf"] = 3.552457
        expected_coef["mitype"] = -0.074266

        assert_array_almost_equal(expected_coef.values, model.coef_)

    @staticmethod
    @pytest.mark.parametrize(
        "fn,expected_file",
        [("predict_survival_function", CGBOOST_SURV_FILE), ("predict_cumulative_hazard_function", CGBOOST_CUMHAZ_FILE)],
    )
    def test_predict_function(make_whas500, fn, expected_file):
        whas500_data = make_whas500(with_std=False, to_numeric=True)

        model = ComponentwiseGradientBoostingSurvivalAnalysis(n_estimators=100, random_state=0)
        train_x, train_y = whas500_data.x[10:], whas500_data.y[10:]
        model.fit(train_x, train_y)

        test_x = whas500_data.x[:10]
        surv_fn = getattr(model, fn)(test_x)

        times = np.unique(train_y["lenfol"][train_y["fstat"]])
        actual = np.vstack([fn_gb(times) for fn_gb in surv_fn])

        expected = np.loadtxt(expected_file, delimiter=",")

        assert_array_almost_equal(actual, expected)

    @staticmethod
    def test_feature_importances(make_whas500):
        whas500_data = make_whas500(with_std=False, to_numeric=True)

        model = ComponentwiseGradientBoostingSurvivalAnalysis(n_estimators=100, random_state=0)
        model.fit(whas500_data.x, whas500_data.y)

        assert whas500_data.x.shape[1] + 1 == len(model.feature_importances_)

    @staticmethod
    def test_fit_verbose(make_whas500):
        whas500_data = make_whas500(with_std=False, to_numeric=True)

        model = ComponentwiseGradientBoostingSurvivalAnalysis(n_estimators=10, verbose=1, random_state=0)
        model.fit(whas500_data.x, whas500_data.y)

    @staticmethod
    def test_ipcwls_loss(make_whas500):
        whas500_data = make_whas500(with_std=False, to_numeric=True)

        model = ComponentwiseGradientBoostingSurvivalAnalysis(loss="ipcwls", n_estimators=100, random_state=0)
        model.fit(whas500_data.x, whas500_data.y)

        assert model.loss == "ipcwls"

        time_predicted = model.predict(whas500_data.x)
        time_true = whas500_data.y["lenfol"]
        event_true = whas500_data.y["fstat"]

        rmse_all = np.sqrt(mean_squared_error(time_true, time_predicted))
        assert rmse_all == pytest.approx(806.283308322, 1e-7)

        rmse_uncensored = np.sqrt(mean_squared_error(time_true[event_true], time_predicted[event_true]))
        assert rmse_uncensored == pytest.approx(542.884585289, 1e-7)

        cindex = model.score(whas500_data.x, whas500_data.y)
        assert cindex == pytest.approx(0.7773356931, 1e-7)

        with pytest.raises(ValueError, match="`fit` must be called with the loss option set to 'coxph'"):
            model.predict_survival_function(whas500_data.x)

        with pytest.raises(ValueError, match="`fit` must be called with the loss option set to 'coxph'"):
            model.predict_cumulative_hazard_function(whas500_data.x)

    @staticmethod
    def test_squared_loss(make_whas500):
        whas500_data = make_whas500(with_std=False, to_numeric=True)

        model = ComponentwiseGradientBoostingSurvivalAnalysis(loss="squared", n_estimators=100, random_state=0)
        model.fit(whas500_data.x, whas500_data.y)

        assert model.loss == "squared"

        time_predicted = model.predict(whas500_data.x)
        time_true = whas500_data.y["lenfol"]
        event_true = whas500_data.y["fstat"]

        rmse_all = np.sqrt(mean_squared_error(time_true, time_predicted))
        assert rmse_all == pytest.approx(793.6256945839657, 1e-7)

        rmse_uncensored = np.sqrt(mean_squared_error(time_true[event_true], time_predicted[event_true]))
        assert rmse_uncensored == pytest.approx(542.83358120153525, 1e-7)

        cindex = model.score(whas500_data.x, whas500_data.y)
        assert cindex == pytest.approx(0.7777082862, 1e-7)

        with pytest.raises(ValueError, match="`fit` must be called with the loss option set to 'coxph'"):
            model.predict_survival_function(whas500_data.x)

        with pytest.raises(ValueError, match="`fit` must be called with the loss option set to 'coxph'"):
            model.predict_cumulative_hazard_function(whas500_data.x)


@pytest.mark.parametrize(
    "model_cls,loss,dropout_rate",
    list(
        itertools.product(
            [ComponentwiseGradientBoostingSurvivalAnalysis, GradientBoostingSurvivalAnalysis],
            ["coxph", "ipcwls", "squared"],
            [0.0, 0.05],
        )
    ),
)
def test_fit_warm_start(make_whas500, model_cls, loss, dropout_rate):
    whas500_data = make_whas500(with_std=False, to_numeric=True)

    model_1 = model_cls(loss=loss, n_estimators=100, learning_rate=0.1, dropout_rate=dropout_rate, random_state=0)
    model_1.fit(whas500_data.x, whas500_data.y)
    assert model_1.n_estimators_ == 100
    pred_1 = model_1.predict(whas500_data.x)

    model_2 = model_cls(loss=loss, n_estimators=40, learning_rate=0.1, dropout_rate=dropout_rate, random_state=0)
    model_2.fit(whas500_data.x, whas500_data.y)
    assert model_2.n_estimators_ == 40
    pred_2_40 = model_2.predict(whas500_data.x)

    model_2.set_params(warm_start=True, n_estimators=100)
    model_2.fit(whas500_data.x, whas500_data.y)
    assert model_2.n_estimators_ == 100
    pred_2 = model_2.predict(whas500_data.x)

    assert_array_almost_equal(pred_1, pred_2)

    # fit from start
    model_2.set_params(warm_start=False, n_estimators=40)
    model_2.fit(whas500_data.x, whas500_data.y)
    assert model_2.n_estimators_ == 40
    assert not hasattr(model_2, "oob_improvement_")

    pred_3 = model_2.predict(whas500_data.x)

    assert_array_almost_equal(pred_2_40, pred_3)

    model_2.set_params(warm_start=True, n_estimators=39)
    msg = r"n_estimators=39 must be larger or equal to estimators_.shape\[0\]=40 when warm_start==True"
    with pytest.raises(ValueError, match=msg):
        model_2.fit(whas500_data.x, whas500_data.y)


@pytest.mark.parametrize(
    "model_cls,subsample",
    list(
        itertools.product(
            [ComponentwiseGradientBoostingSurvivalAnalysis, GradientBoostingSurvivalAnalysis],
            [1.0, 0.666],
        )
    ),
)
def test_fit_warm_start_with_subsample(make_whas500, model_cls, subsample):
    whas500_data = make_whas500(with_std=False, to_numeric=True)

    def assert_subsample_enabled(estimator):
        assert hasattr(estimator, "oob_scores_")
        assert hasattr(estimator, "oob_score_")
        assert hasattr(estimator, "oob_improvement_")

    def assert_subsample_disabled(estimator):
        assert not hasattr(estimator, "oob_scores_")
        assert not hasattr(estimator, "oob_score_")
        assert not hasattr(estimator, "oob_improvement_")

    if subsample < 1:
        assert_model_fit_orig = assert_subsample_enabled
        subsample_alt = 1.0
        assert_model_fit_warm = assert_subsample_enabled  # warm start keeps attributes
        assert_model_fit_alt = assert_subsample_disabled
    else:
        assert_model_fit_orig = assert_subsample_disabled
        subsample_alt = 0.5
        assert_model_fit_warm = assert_subsample_enabled  # adds new attributes
        assert_model_fit_alt = assert_subsample_enabled

    model_1 = model_cls(
        n_estimators=11,
        learning_rate=0.1,
        subsample=subsample,
        random_state=0,
    )
    model_1.fit(whas500_data.x, whas500_data.y)
    assert model_1.n_estimators_ == 11
    assert_model_fit_orig(model_1)

    model_1.set_params(warm_start=True, n_estimators=23)
    model_1.fit(whas500_data.x, whas500_data.y)
    assert model_1.n_estimators_ == 23
    assert_model_fit_orig(model_1)

    model_1.set_params(warm_start=True, n_estimators=50, subsample=subsample_alt)
    model_1.fit(whas500_data.x, whas500_data.y)
    assert model_1.n_estimators_ == 50
    assert_model_fit_warm(model_1)

    model_1.set_params(warm_start=False, n_estimators=25, subsample=subsample_alt)
    model_1.fit(whas500_data.x, whas500_data.y)
    assert model_1.n_estimators_ == 25
    assert_model_fit_alt(model_1)


@pytest.mark.parametrize("model_cls", [ComponentwiseGradientBoostingSurvivalAnalysis, GradientBoostingSurvivalAnalysis])
def test_fit_warm_start_with_wrong_dropout(make_whas500, model_cls):
    whas500_data = make_whas500(with_std=False, to_numeric=True)

    model = model_cls(
        n_estimators=11,
        learning_rate=0.1,
        dropout_rate=0,
        random_state=0,
    )
    model.fit(whas500_data.x, whas500_data.y)
    assert model.n_estimators_ == 11
    assert not hasattr(model, "_scale")

    model.set_params(warm_start=True, n_estimators=50, dropout_rate=0.5)
    msg = (
        "fitting with warm_start=True and dropout_rate > 0 is only "
        "supported if the previous fit used dropout_rate > 0 too"
    )
    with pytest.raises(ValueError, match=msg):
        model.fit(whas500_data.x, whas500_data.y)

    model.set_params(warm_start=False, n_estimators=2)
    model.fit(whas500_data.x, whas500_data.y)
    assert model.n_estimators_ == 2
    assert_array_almost_equal(model._scale, np.array([0.5, 0.5]))


@pytest.fixture(params=[GradientBoostingSurvivalAnalysis, ComponentwiseGradientBoostingSurvivalAnalysis])
def sample_gb_class(request):
    x = np.arange(100).reshape(5, 20)
    y = Surv.from_arrays([False, False, True, True, False], [12, 14, 6, 9, 1])

    return request.param, x, y


@pytest.mark.parametrize("n_estimators", [0, -1])
def test_param_n_estimators(sample_gb_class, n_estimators):
    est_cls, x, y = sample_gb_class
    model = est_cls(n_estimators=n_estimators)

    msg = (
        f"The 'n_estimators' parameter of {est_cls.__name__} must be an int "
        r"in the range \[1, inf\)\. "
        f"Got {n_estimators} instead\\."
    )

    with pytest.raises(ValueError, match=msg):
        model.fit(x, y)


@pytest.mark.parametrize("learning_rate", [-math.ulp(1.0), -1])
def test_param_learning_rate(sample_gb_class, learning_rate):
    est_cls, x, y = sample_gb_class
    model = est_cls(learning_rate=learning_rate)

    msg = (
        f"The 'learning_rate' parameter of {est_cls.__name__} must be a float "
        r"in the range \[0\.0, inf\)\. "
        f"Got {learning_rate} instead\\."
    )

    with pytest.raises(ValueError, match=msg):
        model.fit(x, y)


@pytest.mark.parametrize("subsample", [0, 1.2])
def test_param_subsample(sample_gb_class, subsample):
    est_cls, x, y = sample_gb_class
    model = est_cls(subsample=subsample)

    msg = (
        f"The 'subsample' parameter of {est_cls.__name__} must be a float "
        r"in the range \(0\.0, 1\.0\]\. "
        f"Got {subsample} instead\\."
    )

    with pytest.raises(ValueError, match=msg):
        model.fit(x, y)


@pytest.mark.parametrize("dropout_rate", [-0.1, 1.0, 1.2])
def test_param_dropout_rate(sample_gb_class, dropout_rate):
    est_cls, x, y = sample_gb_class
    model = est_cls(dropout_rate=dropout_rate)

    msg = (
        f"The 'dropout_rate' parameter of {est_cls.__name__} must be a float "
        r"in the range \[0\.0, 1\.0\)\. "
        f"Got {dropout_rate} instead\\."
    )

    with pytest.raises(ValueError, match=msg):
        model.fit(x, y)


def test_param_sample_weight(sample_gb_class):
    est_cls, x, y = sample_gb_class
    model = est_cls()

    with pytest.raises(ValueError, match=r"sample_weight.shape == \(3,\), expected \(5,\)!"):
        model.fit(x, y, [2, 3, 4])

    with pytest.raises(ValueError, match=r"sample_weight.shape == \(8,\), expected \(5,\)!"):
        model.fit(x, y, [2, 4, 5, 6, 7, 1, 2, 7])


@pytest.mark.parametrize("loss", ["", "unknown", None])
def test_param_loss(sample_gb_class, loss):
    est_cls, x, y = sample_gb_class
    model = est_cls(loss=loss)

    msg = rf"The 'loss' parameter of {est_cls.__name__} must be a str among {{.+}}\. Got {loss!r} instead\."

    with pytest.raises(ValueError, match=msg):
        model.fit(x, y)
